/* This software is licensed by the MIT License, see LICENSE file */
/*                                Copyright © 2022 Gregory Lirent */

#include <assert.h>

#include "include.h"

#ifndef LIBCJSONP_SRC_BUFFER_H
#define LIBCJSONP_SRC_BUFFER_H

#define buffer_block 1024

typedef struct {
    char* mem;
    char* ptr;
    size_t  n;
    size_t  c;
} buffer_t;

static_assert(offsetof(vtype_string, buffer) == offsetof(buffer_t, mem), "Implementation assert");

#define buffer_write_and_wrap libcjsonp_builtin_buffer_write_and_wrap
#define buffer_check_size     libcjsonp_builtin_buffer_check_size
#define buffer_write_indent   libcjsonp_builtin_buffer_write_indent
#define buffer_write_char     libcjsonp_builtin_buffer_write_char
#define buffer_write          libcjsonp_builtin_buffer_write

extern void libcjsonp_builtin_buffer_write_and_wrap(buffer_t* x, const char* s);
inline void libcjsonp_builtin_buffer_check_size    (buffer_t* x, size_t need)                    Always_inline__;
inline void libcjsonp_builtin_buffer_write_indent  (buffer_t* x, indent_t* indent, size_t level) Always_inline__;
inline void libcjsonp_builtin_buffer_write_char    (buffer_t* x, char c)                         Always_inline__;
inline void libcjsonp_builtin_buffer_write         (buffer_t* x, const char* v)                  Always_inline__;

inline void libcjsonp_builtin_buffer_check_size(buffer_t* x, size_t need) {
    while (x->c < need) {
        x->mem = libcdsb_realloc(x->mem, x->n + buffer_block);
        x->ptr = x->mem + (x->n - x->c);
        x->n  += buffer_block;
        x->c  += buffer_block;
    }
}

inline void libcjsonp_builtin_buffer_write_char(buffer_t* x, char c) {

    buffer_check_size(x, 1);

    *(x->ptr++) = c;
    --x->c;
}

inline void libcjsonp_builtin_buffer_write(buffer_t* x, const char* v) {
    size_t n = libcdsb_strlen(v);

    buffer_check_size(x, n);

    x->ptr = memcpy(x->ptr, v, n) + n;
    x->c  -= n;
}

inline void libcjsonp_builtin_buffer_write_indent(buffer_t* x, indent_t* indent, size_t level) {

    libcjsonp_builtin_buffer_check_size(x, indent->nmemb * level);

    while (level--) {
        memcpy(x->ptr, indent->indent, indent->nmemb);
        x->ptr += indent->nmemb;
        x->c   -= indent->nmemb;
    }
}

#endif /* LIBCJSONP_SRC_BUFFER_H */
